use crate::sidebar::models::SidebarSelection;
use crate::{content_page::ArticleListMode, settings::SettingsError};
use news_flash::models::ArticleID;
use serde::{Deserialize, Serialize};

static JSON_NAME: &str = "content_page_state.json";

const fn _default_true() -> bool {
    true
}

#[derive(Clone, Default, Debug, Serialize, Deserialize)]
pub struct ContentPageState {
    sidebar_selection: SidebarSelection,
    article_list_mode: ArticleListMode,
    search_term: Option<String>,
    prefer_scraped_content: bool,
    visible_article_id: Option<ArticleID>,
    window_size: (i32, i32),
    maximized: bool,
    article_view_visible: Option<bool>,
    article_view_zoom: Option<f64>,
    #[serde(default = "_default_true")]
    subscriptions_expanded: bool,
    #[serde(default = "_default_true")]
    tags_expanded: bool,

    #[serde(skip)]
    offline: bool,
    #[serde(skip)]
    was_online: bool,
    #[serde(skip)]
    ongoing_article_scrap: bool,
    #[serde(skip)]
    ongoing_sync: bool,
    #[serde(skip)]
    search_focused: bool,
    #[serde(skip)]
    dialog_open: bool,
    #[serde(skip)]
    content_page_visible: bool,
}

const ARTICLE_LIST_PAGE_SIZE: i64 = 20;

impl ContentPageState {
    pub fn new_from_file() -> Result<Self, SettingsError> {
        let path = crate::app::CONFIG_DIR.join(JSON_NAME);
        let data = std::fs::read_to_string(path)?;
        let state = serde_json::from_str::<Self>(&data)?;
        Ok(state)
    }

    pub fn load_from_file(&mut self) -> Result<(), SettingsError> {
        let path = crate::app::CONFIG_DIR.join(JSON_NAME);
        if path.as_path().exists() {
            let data = std::fs::read_to_string(&path)?;
            let state = serde_json::from_str::<Self>(&data)?;

            self.sidebar_selection = state.sidebar_selection;
            self.article_list_mode = state.article_list_mode;
            self.search_term = state.search_term;
            self.prefer_scraped_content = state.prefer_scraped_content;
            self.visible_article_id = state.visible_article_id;
            self.window_size = state.window_size;
            self.maximized = state.maximized;
            self.article_view_visible = state.article_view_visible;
            self.article_view_zoom = state.article_view_zoom;
            self.subscriptions_expanded = state.subscriptions_expanded;
            self.tags_expanded = state.tags_expanded;
        }

        Ok(())
    }

    pub fn write(&self) -> Result<(), SettingsError> {
        let path = crate::app::CONFIG_DIR.join(JSON_NAME);
        let data = serde_json::to_string_pretty(self)?;
        std::fs::write(path, data)?;
        Ok(())
    }

    pub fn set_window_size(&mut self, size: (i32, i32)) {
        self.window_size = size;
    }

    pub fn get_window_size(&self) -> (i32, i32) {
        self.window_size
    }

    pub fn set_maximized(&mut self, maximized: bool) {
        self.maximized = maximized;
    }

    pub fn get_maximized(&self) -> bool {
        self.maximized
    }

    pub fn page_size() -> i64 {
        ARTICLE_LIST_PAGE_SIZE
    }

    pub fn get_sidebar_selection(&self) -> &SidebarSelection {
        &self.sidebar_selection
    }

    pub fn set_sidebar_selection(&mut self, selection: SidebarSelection) {
        self.sidebar_selection = selection;
    }

    pub fn get_article_list_mode(&self) -> ArticleListMode {
        self.article_list_mode
    }

    pub fn set_article_list_mode(&mut self, header: ArticleListMode) {
        self.article_list_mode = header;
    }

    pub fn get_search_term(&self) -> Option<&str> {
        self.search_term.as_deref()
    }

    pub fn set_search_term(&mut self, search_term: Option<String>) {
        self.search_term = search_term;
    }

    pub fn set_offline(&mut self, offline: bool) {
        self.offline = offline;

        if !offline {
            self.was_online = true;
        }
    }

    pub fn get_offline(&self) -> bool {
        self.offline
    }

    pub fn was_online(&self) -> bool {
        self.was_online
    }

    pub fn set_visible_article_id(&mut self, id: Option<&ArticleID>) {
        self.visible_article_id = id.cloned();
    }

    pub fn get_visible_article_id(&self) -> Option<ArticleID> {
        self.visible_article_id.clone()
    }

    pub fn set_prefer_scraped_content(&mut self, pref: bool) {
        self.prefer_scraped_content = pref;
    }

    pub fn get_prefer_scraped_content(&self) -> bool {
        self.prefer_scraped_content
    }

    pub fn started_scraping_article(&mut self) {
        self.ongoing_article_scrap = true;
    }

    pub fn finished_scraping_article(&mut self) {
        self.ongoing_article_scrap = false;
    }

    pub fn is_article_scrap_ongoing(&self) -> bool {
        self.ongoing_article_scrap
    }

    pub fn started_sync(&mut self) {
        self.ongoing_sync = true;
    }

    pub fn finished_sync(&mut self) {
        self.ongoing_sync = false;
    }

    pub fn is_syncing(&self) -> bool {
        self.ongoing_sync
    }

    pub fn get_article_view_visible(&mut self) -> Option<bool> {
        self.article_view_visible.take()
    }

    pub fn set_article_view_visible(&mut self, visible: bool) {
        self.article_view_visible.replace(visible);
    }

    pub fn get_article_view_zoom(&mut self) -> Option<f64> {
        self.article_view_zoom.take()
    }

    pub fn set_article_view_zoom(&mut self, zoom: f64) {
        self.article_view_zoom.replace(zoom);
    }

    pub fn set_subscriptions_expanded(&mut self, expanded: bool) {
        self.subscriptions_expanded = expanded;
    }

    pub fn get_subscriptions_expanded(&self) -> bool {
        self.subscriptions_expanded
    }

    pub fn set_tags_expanded(&mut self, expanded: bool) {
        self.tags_expanded = expanded;
    }

    pub fn get_tags_expanded(&self) -> bool {
        self.tags_expanded
    }

    pub fn set_search_focused(&mut self, focused: bool) {
        self.search_focused = focused;
    }

    pub fn get_search_focused(&self) -> bool {
        self.search_focused
    }

    pub fn set_content_page_visible(&mut self, visible: bool) {
        self.content_page_visible = visible;
    }

    pub fn get_content_page_visible(&self) -> bool {
        self.content_page_visible
    }

    pub fn set_dialog_open(&mut self, open: bool) {
        self.dialog_open = open;
    }

    pub fn get_dialog_open(&self) -> bool {
        self.dialog_open
    }
}
